import os
import threading
import tkinter as tk
from tkinter import ttk

import vlc

from turcar.ui_utils import CommonDialog


class VideoDialog(CommonDialog):
    def __init__(self, master):
        super().__init__(master=master)
        self.title("视频播放器")
        self.geometry("900x650")

        self.protocol("WM_DELETE_WINDOW", self.on_closing)

        self.instance = vlc.Instance("--no-xlib")
        self.player = self.instance.media_player_new()

        self.media = None
        self.media_path = ""
        self.playing = False
        self.display_volume = False

        self.current_time = "00:00"

        self.total_time = "00:00"

        self.playback_speed = 1.0  # 默认倍速为1x

        self.playing_lock = threading.Lock()

        self.style = ttk.Style()

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=0)

        self.video_frame = ttk.Frame(self, borderwidth=1, relief="solid")
        self.bottom_frame = ttk.Frame(self, borderwidth=1, relief="solid")
        self.tree_frame = ttk.Frame(self)

        self.video_frame.place(relx=0, rely=0, relwidth=0.75, relheight=0.75)
        self.bottom_frame.place(relx=0, rely=0.75, relwidth=0.75, relheight=0.25)
        self.tree_frame.place(relx=0.75, rely=0, relwidth=0.25, relheight=1)

        self.tree_frame.rowconfigure(0, weight=1)  # 设置第0行的权重，以填充垂直空间

        self.vlc_player = self.instance.media_player_new()
        self.vlc_player.set_fullscreen(False)
        self.vlc_player.set_hwnd(self.video_frame.winfo_id())
        self.vlc_player.audio_set_volume(50)

        # 图像引入
        img_dir = os.path.join(os.path.dirname(__file__), "res")
        self.video_play = tk.PhotoImage("video_play", file=os.path.join(img_dir, "video-play.png"))

        self.video_stop = tk.PhotoImage("video_stop", file=os.path.join(img_dir, "video-stop.png"))

        self.video_rewind = tk.PhotoImage(
            "video_rewind", file=os.path.join(img_dir, "video-rewind.png")
        )

        self.video_fast_forward = tk.PhotoImage(
            "video_fast_forward", file=os.path.join(img_dir, "video-fast-forward.png")
        )

        self.video_mute = tk.PhotoImage("video_mute", file=os.path.join(img_dir, "video-mute.png"))

        self.video_voice = tk.PhotoImage(
            "video_voice", file=os.path.join(img_dir, "video-voice.png")
        )

        self.current_time_label = tk.Label(self.bottom_frame, text=self.current_time)
        self.current_time_label.place(relx=0.20, rely=0.2, relheight=0.1, relwidth=0.06)

        self.progress_var = tk.DoubleVar()
        self.progress_bar = ttk.Progressbar(
            self.bottom_frame, variable=self.progress_var, maximum=100, mode="determinate"
        )
        self.progress_bar.place(relx=0.27, rely=0.24, relheight=0.04, relwidth=0.46)
        self.progress_bar.bind("<Button-1>", self.seek_to_progress)
        self.progress_bar["style"] = "TProgressbar"

        self.progress_bar.rowconfigure(0, weight=1)  # 设置第0行的权重，以填充垂直空间

        self.total_time_label = tk.Label(self.bottom_frame, text=self.total_time, anchor="w")
        self.total_time_label.place(relx=0.74, rely=0.2, relheight=0.1, relwidth=0.062)

        self.playback_speed_var = tk.StringVar(value=str(self.playback_speed) + "x")

        self.playback_speed_menu = ttk.Combobox(
            self.bottom_frame,
            textvariable=self.playback_speed_var,
            values=["0.5x", "1.0x", "2.0x"],
            state="readonly",
        )
        self.playback_speed_menu.place(relx=0.12, rely=0.51, relheight=0.22, relwidth=0.1)
        self.playback_speed_menu.bind("<<ComboboxSelected>>", self.update_playback_speed)

        self.rewind_button = ttk.Button(
            self.bottom_frame, image=self.video_rewind, command=self.rewind
        )
        self.rewind_button.place(relx=0.3, rely=0.5, relheight=0.25, relwidth=0.1)

        self.play_pause_button = ttk.Button(
            self.bottom_frame, image=self.video_play, command=self.play_pause
        )
        self.play_pause_button.place(relx=0.45, rely=0.5, relheight=0.25, relwidth=0.1)

        self.fast_forward_button = ttk.Button(
            self.bottom_frame, image=self.video_fast_forward, command=self.fast_forward
        )
        self.fast_forward_button.place(relx=0.6, rely=0.5, relheight=0.25, relwidth=0.1)

        self.voice_button = ttk.Button(
            self.bottom_frame, image=self.video_voice, command=self.show_volume
        )
        self.voice_button.place(relx=0.77, rely=0.5, relheight=0.25, relwidth=0.1)
        # 音量滑块
        self.volume_slider = ttk.Scale(
            self.bottom_frame, from_=0, to=100, orient="horizontal", command=self.update_volume
        )
        self.volume_slider.set(50)  # 初始音量为50
        self.volume_slider.place_forget()

        self.tree = ttk.Treeview(self.tree_frame, columns=("Type", "Path"))
        self.tree.heading("#0", text="Name")
        self.tree.heading("#1", text="Type")
        self.tree.heading("#2", text="Path")
        self.tree.column("#1", width=0)  # 隐藏第一列
        self.tree.column("#2", width=0)  # 隐藏第二列
        self.tree.place(relx=0, rely=0, relwidth=0.9, relheight=1)

        self.scrollbar = ttk.Scrollbar(self.tree_frame, orient="vertical", command=self.tree.yview)
        self.scrollbar.place(relx=0.9, rely=0, relwidth=0.1, relheight=1)

        self.tree.configure(yscrollcommand=self.scrollbar.set)

        file_path = "../video"
        if not os.path.exists(file_path):
            os.makedirs(file_path)

        self.populate_tree("", file_path)  # 替换为实际的文件夹路径

        self.tree.bind("<Double-1>", self.on_tree_double_click)

        self.playback_index = 0

    def show_volume(self):
        if not self.display_volume:
            self.volume_slider.place(relx=0.67, rely=0.4, relheight=0.05, relwidth=0.3)
        else:
            self.volume_slider.place_forget()
        self.display_volume = not self.display_volume

    def update_volume(self, volume):
        volume = int(float(volume))
        self.vlc_player.audio_set_volume(volume)

    def update_playback_speed(self, event):
        speed_str = self.playback_speed_var.get()
        speed = float(speed_str.replace("x", ""))
        self.playback_speed = speed

        if self.vlc_player.get_rate() != self.playback_speed:
            self.vlc_player.set_rate(self.playback_speed)

    def seek_to_progress(self, event):
        if self.media is not None:
            progress_width = self.progress_bar.winfo_width()
            click_x = event.x
            seek_time = int((click_x / progress_width) * self.vlc_player.get_length())
            self.vlc_player.set_time(seek_time)

    def populate_tree(self, parent, path):
        for item in os.listdir(path):
            item_path = os.path.join(path, item)
            if os.path.isdir(item_path):
                node = self.tree.insert(parent, "end", text=item, values=("folder", item_path))
                self.populate_tree(node, item_path)
            elif item.lower().endswith((".mp4", ".avi", ".mkv")):
                self.tree.insert(parent, "end", text=item, values=("file", item_path))

    def on_tree_double_click(self, event):
        item = self.tree.selection()[0]
        item_type = self.tree.set(item, "#1")

        if item_type == "file":
            self.media_path = self.tree.set(item, "#2")
            if self.media_path.lower().endswith((".mp4", ".avi", ".mkv")):
                self.playing = False
                self.media = self.instance.media_new(self.media_path)
                self.vlc_player.set_media(self.media)
                self.play_pause()

    def play_pause(self):
        if self.media is not None:
            if self.playing:
                self.vlc_player.pause()
                self.play_pause_button.config(image=self.video_play)
                self.playing = False  # 停止播放状态
            else:
                self.vlc_player.play()
                self.play_pause_button.config(image=self.video_stop)
                self.after(500, self.update_labels_in_main_thread)
                self.playing = True  # 标记为播放状态

    def rewind(self):
        if self.media is not None:
            current_time = self.vlc_player.get_time()
            self.vlc_player.set_time(max(0, current_time - 10000))

    def fast_forward(self):
        if self.media is not None:
            current_time = self.vlc_player.get_time()
            self.vlc_player.set_time(current_time + 10000)

    def play_next(self):
        selected_item = self.tree.selection()
        next_item = self.tree.next(selected_item[0])
        while next_item:
            item_type = self.tree.set(next_item, "#1")
            if item_type == "file":
                self.playing = False
                self.tree.selection_set(next_item)
                self.media_path = self.tree.set(next_item, "#2")
                self.media = self.instance.media_new(self.media_path)
                self.vlc_player.set_media(self.media)
                self.play_pause()  # Directly call play_pause to play the next video
                return
            next_item = self.tree.next(next_item)

        # If no more file items are found, stop playback
        self.playing = False
        self.play_pause_button.config(image=self.video_play)
        self.progress_var.set(0)
        self.media = None  # Reset the media instance after all videos have been played

    def update_labels_in_main_thread(self):
        if self.playing:
            current_time = self.vlc_player.get_time()
            total_time = self.vlc_player.get_length()
            current_time_str = self.format_time(current_time)
            total_time_str = self.format_time(total_time)
            self.current_time_label.config(text=current_time_str)
            self.total_time_label.config(text=total_time_str)
            self.after(500, self.update_labels_in_main_thread)  # 再次调度更新

            if total_time > 0:
                progress = (current_time / total_time) * 100
                self.progress_var.set(progress)
            if current_time >= total_time:
                self.play_next()

    def format_time(self, milliseconds):
        total_seconds = int(milliseconds / 1000)
        minutes = total_seconds // 60
        seconds = total_seconds % 60
        return f"{minutes:02d}:{seconds:02d}"

    def on_closing(self):
        if self.playing:
            self.vlc_player.stop()  # 停止音频播放
            self.playing = False  # 停止播放线程的执行
        self.destroy()  # 销毁窗口



"""
control of WS2812 / NeoPixel LEDs.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/neopixel.rst.
=====================================================

.. module:: neopixel
   :synopsis: control of WS2812 / NeoPixel LEDs

This module provides a driver for WS2818 / NeoPixel LEDs.

.. note:: This module is only included by default on the ESP8266 and ESP32
   ports. On STM32 / Pyboard, you can `download the module
   <https://github.com/micropython/micropython/blob/master/drivers/neopixel/neopixel.py>`_
   and copy it to the filesystem.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final

from machine import Pin

_Color: Final = tuple[int, int, int] | tuple[int, int, int, int]

class NeoPixel:
    """
   This class stores pixel data for a WS2812 LED strip connected to a pin. The
   application should set pixel data and then call :meth:`NeoPixel.write`
   when it is ready to update the strip.
   
   For example::
   
       import neopixel
   
       # 32 LED strip connected to X8.
       p = machine.Pin.board.X8
       n = neopixel.NeoPixel(p, 32)
   
       # Draw a red gradient.
       for i in range(32):
           n[i] = (i * 8, 0, 0)
   
       # Update the strip.
       n.write()
   """

    def __init__(self, pin: Pin, n: int, /, *, bpp: int = 3, timing: int = 1):
        """
       Construct an NeoPixel object.  The parameters are:
       
           - *pin* is a machine.Pin instance.
           - *n* is the number of LEDs in the strip.
           - *bpp* is 3 for RGB LEDs, and 4 for RGBW LEDs.
           - *timing* is 0 for 400KHz, and 1 for 800kHz LEDs (most are 800kHz).
      """
    def fill(self, pixel: _Color, /) -> None:
        """
       Sets the value of all pixels to the specified *pixel* value (i.e. an
       RGB/RGBW tuple).
      """
    def __len__(self) -> int:
        """
       Returns the number of LEDs in the strip.
      """
    def __setitem__(self, index: int, val: _Color, /) -> None:
        """
       Set the pixel at *index* to the value, which is an RGB/RGBW tuple.
      """
    def __getitem__(self, index: int, /) -> _Color:
        """
       Returns the pixel at *index* as an RGB/RGBW tuple.
      """
    def write(self) -> None:
        """
       Writes the current pixel data to the strip.
      """



"""
regular expressions.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/re.rst.
=======================================

.. module:: re
   :synopsis: regular expressions

|see_cpython_module| :mod:`python:re`.

This module implements regular expression operations. Regular expression
syntax supported is a subset of CPython ``re`` module (and actually is
a subset of POSIX extended regular expressions).

Supported operators and special sequences are:

``.``
   Match any character.

``[...]``
   Match set of characters. Individual characters and ranges are supported,
   including negated sets (e.g. ``[^a-c]``).

``^``
   Match the start of the string.

``$``
   Match the end of the string.

``?``
   Match zero or one of the previous sub-pattern.

``*``
   Match zero or more of the previous sub-pattern.

``+``
   Match one or more of the previous sub-pattern.

``??``
   Non-greedy version of ``?``, match zero or one, with the preference
   for zero.

``*?``
   Non-greedy version of ``*``, match zero or more, with the preference
   for the shortest match.

``+?``
   Non-greedy version of ``+``, match one or more, with the preference
   for the shortest match.

``|``
   Match either the left-hand side or the right-hand side sub-patterns of
   this operator.

``(...)``
   Grouping. Each group is capturing (a substring it captures can be accessed
   with `match.group()` method).

``\d``
   Matches digit. Equivalent to ``[0-9]``.

``\D``
   Matches non-digit. Equivalent to ``[^0-9]``.

``\s``
   Matches whitespace. Equivalent to ``[ \t-\r]``.

``\S``
   Matches non-whitespace. Equivalent to ``[^ \t-\r]``.

``\w``
   Matches "word characters" (ASCII only). Equivalent to ``[A-Za-z0-9_]``.

``\W``
   Matches non "word characters" (ASCII only). Equivalent to ``[^A-Za-z0-9_]``.

``\``
   Escape character. Any other character following the backslash, except
   for those listed above, is taken literally. For example, ``\*`` is
   equivalent to literal ``*`` (not treated as the ``*`` operator).
   Note that ``\r``, ``\n``, etc. are not handled specially, and will be
   equivalent to literal letters ``r``, ``n``, etc. Due to this, it's
   not recommended to use raw Python strings (``r""``) for regular
   expressions. For example, ``r"\r\n"`` when used as the regular
   expression is equivalent to ``"rn"``. To match CR character followed
   by LF, use ``"\r\n"``.

**NOT SUPPORTED**:

* counted repetitions (``{m,n}``)
* named groups (``(?P<name>...)``)
* non-capturing groups (``(?:...)``)
* more advanced assertions (``\b``, ``\B``)
* special character escapes like ``\r``, ``\n`` - use Python's own escaping
  instead
* etc.

Example::

    import re

    # As re doesn't support escapes itself, use of r"" strings is not
    # recommended.
    regex = re.compile("[\r\n]")

    regex.split("line1\rline2\nline3\r\n")

    # Result:
    # ['line1', 'line2', 'line3', '', '']
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import AnyStr, Callable, Generic, Final, Any

_StrLike: Final = str | bytes

def compile(regex_str: _StrLike, flags: int = ..., /) -> "ure":
    """
   Compile regular expression, return `regex <regex>` object.
   """

def match(regex_str: _StrLike, string: AnyStr, /) -> "Match[AnyStr]":
    """
   Compile *regex_str* and match against *string*. Match always happens
   from starting position in a string.
   """

def search(regex_str: _StrLike, string: AnyStr, /) -> "Match[AnyStr]":
    """
   Compile *regex_str* and search it in a *string*. Unlike `match`, this will search
   string for first position which matches regex (which still may be
   0 if regex is anchored).
   """

def sub(
    regex_str: _StrLike,
    replace: AnyStr | Callable[["Match[AnyStr]"], AnyStr],
    string: AnyStr,
    count: int = 0,
    flags: int = 0,
    /,
) -> AnyStr:
    """
   Compile *regex_str* and search for it in *string*, replacing all matches
   with *replace*, and returning the new string.
   
   *replace* can be a string or a function.  If it is a string then escape
   sequences of the form ``\<number>`` and ``\g<number>`` can be used to
   expand to the corresponding group (or an empty string for unmatched groups).
   If *replace* is a function then it must take a single argument (the match)
   and should return a replacement string.
   
   If *count* is specified and non-zero then substitution will stop after
   this many substitutions are made.  The *flags* argument is ignored.
   
   Note: availability of this function depends on :term:`MicroPython port`.
   """

DEBUG: Final[int] = ...
"""
Flag value, display debug information about compiled expression.
   (Availability depends on :term:`MicroPython port`.)
"""

# noinspection PyPep8Naming
class ure:
    """
   Compiled regular expression. Instances of this class are created using
   `re.compile()`.
   """

    def match(self, string: AnyStr, /) -> "Match[AnyStr]":
        """
      Similar to the module-level functions :meth:`match`, :meth:`search`
      and :meth:`sub`.
      Using methods is (much) more efficient if the same regex is applied to
      multiple strings.
      """
    def search(self, string: AnyStr, /) -> "Match[AnyStr]":
        """
      Similar to the module-level functions :meth:`match`, :meth:`search`
      and :meth:`sub`.
      Using methods is (much) more efficient if the same regex is applied to
      multiple strings.
      """
    def sub(
        self,
        replace: AnyStr | Callable[["Match[AnyStr]"], AnyStr],
        string: AnyStr,
        count: int = 0,
        flags: int = 0,
        /,
    ) -> AnyStr:
        """
      Similar to the module-level functions :meth:`match`, :meth:`search`
      and :meth:`sub`.
      Using methods is (much) more efficient if the same regex is applied to
      multiple strings.
      """
    def split(self, string: AnyStr, max_split: int = -1, /) -> list[AnyStr]:
        """
      Split a *string* using regex. If *max_split* is given, it specifies
      maximum number of splits to perform. Returns list of strings (there
      may be up to *max_split+1* elements if it's specified).
      """

class Match(Generic[AnyStr]):
    """
   Match objects as returned by `match()` and `search()` methods, and passed
   to the replacement function in `sub()`.
   
   The name, `Match`, used for typing is not the same as the runtime name, `match` (note lowercase `m`).
   The reason for this difference is that the runtime uses `match` as both a class name and as a method name and
   this is not possible within code written entirely in Python and therefore not possible within typing code.
   """

    def group(self, index: int, /) -> AnyStr:
        """
      Return matching (sub)string. *index* is 0 for entire match,
      1 and above for each capturing group. Only numeric groups are supported.
      """
    def groups(self) -> tuple[AnyStr | Any, ...]:
        """
      Return a tuple containing all the substrings of the groups of the match.
      
      Note: availability of this method depends on :term:`MicroPython port`.
      """
    def start(self, index: int = ..., /) -> int:
        """
      Return the index in the original string of the start or end of the
      substring group that was matched.  *index* defaults to the entire
      group, otherwise it will select a group.
      
      Note: availability of these methods depends on :term:`MicroPython port`.
      """
    def end(self, index: int = ..., /) -> int:
        """
      Return the index in the original string of the start or end of the
      substring group that was matched.  *index* defaults to the entire
      group, otherwise it will select a group.
      
      Note: availability of these methods depends on :term:`MicroPython port`.
      """
    def span(self, index: int = ..., /) -> tuple[int, int]:
        """
      Returns the 2-tuple ``(match.start(index), match.end(index))``.
      
      Note: availability of this method depends on :term:`MicroPython port`.
      """



"""
socket.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/socket.rst.
******************************

.. module:: socket
   :synopsis: socket module

|see_cpython_module| :mod:`python:socket`.

This module provides access to the BSD socket interface.

.. admonition:: Difference to CPython
   :class: attention

   For efficiency and consistency, socket objects in MicroPython implement a `stream`
   (file-like) interface directly. In CPython, you need to convert a socket to
   a file-like object using `makefile()` method. This method is still supported
   by MicroPython (but is a no-op), so where compatibility with CPython matters,
   be sure to use it.

Socket address format(s)
------------------------

The native socket address format of the ``socket`` module is an opaque data type
returned by `getaddrinfo` function, which must be used to resolve textual address
(including numeric addresses)::

    sockaddr = socket.getaddrinfo('www.micropython.org', 80)[0][-1]
    # You must use getaddrinfo() even for numeric addresses
    sockaddr = socket.getaddrinfo('127.0.0.1', 80)[0][-1]
    # Now you can use that address
    sock.connect(addr)

Using `getaddrinfo` is the most efficient (both in terms of memory and processing
power) and portable way to work with addresses.

However, ``socket`` module (note the difference with native MicroPython
``socket`` module described here) provides CPython-compatible way to specify
addresses using tuples, as described below. Note that depending on a
:term:`MicroPython port`, ``socket`` module can be builtin or need to be
installed from `micropython-lib` (as in the case of :term:`MicroPython Unix port`),
and some ports still accept only numeric addresses in the tuple format,
and require to use `getaddrinfo` function to resolve domain names.

Summing up:

* Always use `getaddrinfo` when writing portable applications.
* Tuple addresses described below can be used as a shortcut for
  quick hacks and interactive use, if your port supports them.

Tuple address format for ``socket`` module:

* IPv4: *(ipv4_address, port)*, where *ipv4_address* is a string with
  dot-notation numeric IPv4 address, e.g. ``"8.8.8.8"``, and *port* is and
  integer port number in the range 1-65535. Note the domain names are not
  accepted as *ipv4_address*, they should be resolved first using
  `socket.getaddrinfo()`.
* IPv6: *(ipv6_address, port, flowinfo, scopeid)*, where *ipv6_address*
  is a string with colon-notation numeric IPv6 address, e.g. ``"2001:db8::1"``,
  and *port* is an integer port number in the range 1-65535. *flowinfo*
  must be 0. *scopeid* is the interface scope identifier for link-local
  addresses. Note the domain names are not accepted as *ipv6_address*,
  they should be resolved first using `socket.getaddrinfo()`. Availability
  of IPv6 support depends on a :term:`MicroPython port`.

   

   .. exception:: socket.error
   
      MicroPython does NOT have this exception.
   
      .. admonition:: Difference to CPython
           :class: attention
   
           CPython used to have a ``socket.error`` exception which is now deprecated,
           and is an alias of `OSError`. In MicroPython, use `OSError` directly.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final, Any, Literal, overload

from uio import AnyReadableBuf, AnyWritableBuf

_Address: Final = tuple[str, int] | tuple[str, int, int, int] | str

def socket(
    af: int = "AF_INET", type: int = "SOCK_STREAM", proto: int = "IPPROTO_TCP", /,
) -> "Socket":
    """
   Create a new socket using the given address family, socket type and
   protocol number. Note that specifying *proto* in most cases is not
   required (and not recommended, as some MicroPython ports may omit
   ``IPPROTO_*`` constants). Instead, *type* argument will select needed
   protocol automatically::
   
        # Create STREAM TCP socket
        socket(AF_INET, SOCK_STREAM)
        # Create DGRAM UDP socket
        socket(AF_INET, SOCK_DGRAM)
   """

def getaddrinfo(
    host: str, port: int, af: int = 0, type: int = 0, proto: int = 0, flags: int = 0, /,
) -> list[tuple[int, int, int, str, tuple[str, int] | tuple[str, int, int, int]]]:
    """
   Translate the host/port argument into a sequence of 5-tuples that contain all the
   necessary arguments for creating a socket connected to that service. Arguments
   *af*, *type*, and *proto* (which have the same meaning as for the `socket()` function)
   can be used to filter which kind of addresses are returned. If a parameter is not
   specified or zero, all combinations of addresses can be returned (requiring
   filtering on the user side).
   
   The resulting list of 5-tuples has the following structure::
   
      (family, type, proto, canonname, sockaddr)
   
   The following example shows how to connect to a given url::
   
      s = socket.socket()
      # This assumes that if "type" is not specified, an address for
      # SOCK_STREAM will be returned, which may be not true
      s.connect(socket.getaddrinfo('www.micropython.org', 80)[0][-1])
   
   Recommended use of filtering params::
   
      s = socket.socket()
      # Guaranteed to return an address which can be connect'ed to for
      # stream operation.
      s.connect(socket.getaddrinfo('www.micropython.org', 80, 0, SOCK_STREAM)[0][-1])
   
   .. admonition:: Difference to CPython
      :class: attention
   
      CPython raises a ``socket.gaierror`` exception (`OSError` subclass) in case
      of error in this function. MicroPython doesn't have ``socket.gaierror``
      and raises OSError directly. Note that error numbers of `getaddrinfo()`
      form a separate namespace and may not match error numbers from
      the :mod:`errno` module. To distinguish `getaddrinfo()` errors, they are
      represented by negative numbers, whereas standard system errors are
      positive numbers (error numbers are accessible using ``e.args[0]`` property
      from an exception object). The use of negative values is a provisional
      detail which may change in the future.
   """

def inet_ntop(af: int, bin_addr: bytes, /) -> str:
    """
   Convert a binary network address *bin_addr* of the given address family *af*
   to a textual representation::
   
        >>> socket.inet_ntop(socket.AF_INET, b"\x7f\0\0\1")
        '127.0.0.1'
   """

def inet_pton(af: int, txt_addr: str, /) -> bytes:
    """
   Convert a textual network address *txt_addr* of the given address family *af*
   to a binary representation::
   
        >>> socket.inet_pton(socket.AF_INET, "1.2.3.4")
        b'\x01\x02\x03\x04'
   """

AF_INET: Final[int] = ...
"""
Address family types. Availability depends on a particular :term:`MicroPython port`.
"""

AF_INET6: Final[int] = ...
"""
Address family types. Availability depends on a particular :term:`MicroPython port`.
"""

SOCK_STREAM: Final[int] = ...
"""
Socket types.
"""

SOCK_DGRAM: Final[int] = ...
"""
Socket types.
"""

IPPROTO_UDP: Final[int] = ...
"""
IP protocol numbers. Availability depends on a particular :term:`MicroPython port`.
   Note that you don't need to specify these in a call to `socket.socket()`,
   because `SOCK_STREAM` socket type automatically selects `IPPROTO_TCP`, and
   `SOCK_DGRAM` - `IPPROTO_UDP`. Thus, the only real use of these constants
   is as an argument to `setsockopt()`.
"""

IPPROTO_TCP: Final[int] = ...
"""
IP protocol numbers. Availability depends on a particular :term:`MicroPython port`.
   Note that you don't need to specify these in a call to `socket.socket()`,
   because `SOCK_STREAM` socket type automatically selects `IPPROTO_TCP`, and
   `SOCK_DGRAM` - `IPPROTO_UDP`. Thus, the only real use of these constants
   is as an argument to `setsockopt()`.
"""

SOL_SOCKET: Final[int] = ...
"""
      Socket option levels (an argument to `setsockopt()`). The exact
      inventory depends on a :term:`MicroPython port`.
"""

SO_REUSEADDR: Final[int] = ...
"""
      Socket options (an argument to `setsockopt()`). The exact
      inventory depends on a :term:`MicroPython port`.
"""

IPPROTO_SEC: Final[int] = ...
"""
Special protocol value to create SSL-compatible socket.


   Constants specific to WiPy:
"""

class Socket:
    """
   A unix like socket, for more information see module ``socket``'s description.
   
   The name, `Socket`, used for typing is not the same as the runtime name, `socket` (note lowercase `s`).
   The reason for this difference is that the runtime uses `socket` as both a class name and as a method name and
   this is not possible within code written entirely in Python and therefore not possible within typing code.
   """

    def close(self) -> None:
        """
      Mark the socket closed and release all resources. Once that happens, all future operations
      on the socket object will fail. The remote end will receive EOF indication if
      supported by protocol.
      
      Sockets are automatically closed when they are garbage-collected, but it is recommended
      to `close()` them explicitly as soon you finished working with them.
      """
    def bind(self, address: _Address | bytes, /) -> None:
        """
      Bind the socket to *address*. The socket must not already be bound.
      """
    def listen(self, backlog: int = ..., /) -> None:
        """
      Enable a server to accept connections. If *backlog* is specified, it must be at least 0
      (if it's lower, it will be set to 0); and specifies the number of unaccepted connections
      that the system will allow before refusing new connections. If not specified, a default
      reasonable value is chosen.
      """
    def accept(self) -> None:
        """
      Accept a connection. The socket must be bound to an address and listening for connections.
      The return value is a pair (conn, address) where conn is a new socket object usable to send
      and receive data on the connection, and address is the address bound to the socket on the
      other end of the connection.
      """
    def connect(self, address: _Address | bytes, /) -> None:
        """
      Connect to a remote socket at *address*.
      """
    def send(self, bytes: AnyReadableBuf, /) -> int:
        """
      Send data to the socket. The socket must be connected to a remote socket.
      Returns number of bytes sent, which may be smaller than the length of data
      ("short write").
      """
    def sendall(self, bytes: AnyReadableBuf, /) -> None:
        """
      Send all data to the socket. The socket must be connected to a remote socket.
      Unlike `send()`, this method will try to send all of data, by sending data
      chunk by chunk consecutively.
      
      The behaviour of this method on non-blocking sockets is undefined. Due to this,
      on MicroPython, it's recommended to use `write()` method instead, which
      has the same "no short writes" policy for blocking sockets, and will return
      number of bytes sent on non-blocking sockets.
      """
    def recv(self, bufsize: int, /) -> bytes:
        """
      Receive data from the socket. The return value is a bytes object representing the data
      received. The maximum amount of data to be received at once is specified by bufsize.
      """
    def sendto(self, bytes: AnyReadableBuf, address: _Address, /) -> int:
        """
      Send data to the socket. The socket should not be connected to a remote socket, since the
      destination socket is specified by *address*.
      """
    def recvfrom(self, bufsize: int, /) -> tuple[bytes, Any]:
        """
     Receive data from the socket. The return value is a pair *(bytes, address)* where *bytes* is a
     bytes object representing the data received and *address* is the address of the socket sending
     the data.
      """
    def setsockopt(
        self, level: int, optname: int, value: AnyReadableBuf | int, /
    ) -> None:
        """
      Set the value of the given socket option. The needed symbolic constants are defined in the
      socket module (SO_* etc.). The *value* can be an integer or a bytes-like object representing
      a buffer.
      """
    def settimeout(self, value: float | None, /) -> None:
        """
      **Note**: Not every port supports this method, see below.
      
      Set a timeout on blocking socket operations. The value argument can be a nonnegative floating
      point number expressing seconds, or None. If a non-zero value is given, subsequent socket operations
      will raise an `OSError` exception if the timeout period value has elapsed before the operation has
      completed. If zero is given, the socket is put in non-blocking mode. If None is given, the socket
      is put in blocking mode.
      
      Not every :term:`MicroPython port` supports this method. A more portable and
      generic solution is to use `select.poll` object. This allows to wait on
      multiple objects at the same time (and not just on sockets, but on generic
      `stream` objects which support polling). Example::
      
           # Instead of:
           s.settimeout(1.0)  # time in seconds
           s.read(10)  # may timeout
      
           # Use:
           poller = select.poll()
           poller.register(s, select.POLLIN)
           res = poller.poll(1000)  # time in milliseconds
           if not res:
               # s is still not ready for input, i.e. operation timed out
      
      .. admonition:: Difference to CPython
         :class: attention
      
         CPython raises a ``socket.timeout`` exception in case of timeout,
         which is an `OSError` subclass. MicroPython raises an OSError directly
         instead. If you use ``except OSError:`` to catch the exception,
         your code will work both in MicroPython and CPython.
      """
    def setblocking(self, value: bool, /) -> None:
        """
      Set blocking or non-blocking mode of the socket: if flag is false, the socket is set to non-blocking,
      else to blocking mode.
      
      This method is a shorthand for certain `settimeout()` calls:
      
      * ``sock.setblocking(True)`` is equivalent to ``sock.settimeout(None)``
      * ``sock.setblocking(False)`` is equivalent to ``sock.settimeout(0)``
      """
    @overload
    def makefile(
        self, mode: Literal["rb", "wb", "rwb"] = "rb", buffering: int = 0, /
    ) -> Socket:
        """
      Return a file object associated with the socket. The exact returned type depends on the arguments
      given to makefile(). The support is limited to binary modes only ('rb', 'wb', and 'rwb').
      CPython's arguments: *encoding*, *errors* and *newline* are not supported.
      
      .. admonition:: Difference to CPython
         :class: attention
      
         As MicroPython doesn't support buffered streams, values of *buffering*
         parameter is ignored and treated as if it was 0 (unbuffered).
      
      .. admonition:: Difference to CPython
         :class: attention
      
         Closing the file object returned by makefile() WILL close the
         original socket as well.
      """
    @overload
    def makefile(self, mode: str, buffering: int = 0, /) -> Socket:
        """
      Return a file object associated with the socket. The exact returned type depends on the arguments
      given to makefile(). The support is limited to binary modes only ('rb', 'wb', and 'rwb').
      CPython's arguments: *encoding*, *errors* and *newline* are not supported.
      
      .. admonition:: Difference to CPython
         :class: attention
      
         As MicroPython doesn't support buffered streams, values of *buffering*
         parameter is ignored and treated as if it was 0 (unbuffered).
      
      .. admonition:: Difference to CPython
         :class: attention
      
         Closing the file object returned by makefile() WILL close the
         original socket as well.
      """
    @overload
    def read(self) -> bytes:
        """
      Read up to size bytes from the socket. Return a bytes object. If *size* is not given, it
      reads all data available from the socket until EOF; as such the method will not return until
      the socket is closed. This function tries to read as much data as
      requested (no "short reads"). This may be not possible with
      non-blocking socket though, and then less data will be returned.
      """
    @overload
    def read(self, size: int, /) -> bytes:
        """
      Read up to size bytes from the socket. Return a bytes object. If *size* is not given, it
      reads all data available from the socket until EOF; as such the method will not return until
      the socket is closed. This function tries to read as much data as
      requested (no "short reads"). This may be not possible with
      non-blocking socket though, and then less data will be returned.
      """
    @overload
    def readinto(self, buf: AnyWritableBuf, /) -> int | None:
        """
      Read bytes into the *buf*.  If *nbytes* is specified then read at most
      that many bytes.  Otherwise, read at most *len(buf)* bytes. Just as
      `read()`, this method follows "no short reads" policy.
      
      Return value: number of bytes read and stored into *buf*.
      """
    @overload
    def readinto(self, buf: AnyWritableBuf, nbytes: int, /) -> int | None:
        """
      Read bytes into the *buf*.  If *nbytes* is specified then read at most
      that many bytes.  Otherwise, read at most *len(buf)* bytes. Just as
      `read()`, this method follows "no short reads" policy.
      
      Return value: number of bytes read and stored into *buf*.
      """
    def readline(self) -> bytes:
        """
      Read a line, ending in a newline character.
      
      Return value: the line read.
      """
    def write(self, buf: AnyReadableBuf, /) -> int | None:
        """
      Write the buffer of bytes to the socket. This function will try to
      write all data to a socket (no "short writes"). This may be not possible
      with a non-blocking socket though, and returned value will be less than
      the length of *buf*.
      
      Return value: number of bytes written.
      """



from __future__ import annotations

class CywPin:
    """A class that represents a GPIO pin attached to the wifi chip.

    Cannot be constructed at runtime, but may be the type of a pin object
    in :py:mod:`board`. A `CywPin` can be used as a DigitalInOut, but not with other
    peripherals such as `PWMOut`."""

PM_STANDARD: int
"""The standard power management mode"""
PM_AGGRESSIVE: int
"""Aggressive power management mode for optimal power usage at the cost of performance"""
PM_PERFORMANCE: int
"""Performance power management mode where more power is used to increase performance"""
PM_DISABLED: int
"""Disable power management and always use highest power mode. CircuitPython sets this value at reset time, because it provides the best connectivity reliability."""

def set_power_management(value: int) -> None:
    """Set the power management register

    For transmitter power, see ``wifi.Radio.txpower``.
    This controls software power saving features inside the cyw43 chip.
    it does not control transmitter power.

    The value is interpreted as a 24-bit hexadecimal number of the form
    ``0x00adbrrm``.

    The low 4 bits, ``m``, are the power management mode:
     * 0: disabled
     * 1: aggressive power saving which reduces wifi throughput
     * 2: Power saving with high throughput

    The next 8 bits, ``r``, specify "the maximum time to wait before going back to sleep" for power management mode 2. The units of ``r`` are 10ms.

    The next 4 bits, ``b``, are the "wake period is measured in beacon periods".

    The next 4 bits, ``d``, specify the "wake interval measured in DTIMs. If this is set to 0, the wake interval is measured in beacon periods".

    The top 4 bits, ``a``, specifies the "wake interval sent to the access point"

    Several ``PM_`` constants gathered from various sources are included
    in this module.  According to Raspberry Pi documentation, the value 0xa11140
    (called `cyw43.PM_DISABLED` here) increases responsiveness at the cost of higher power
    usage.
    """

def get_power_management() -> int:
    """Retrieve the power management register"""



class UserError(RuntimeError):
    pass


class CommunicationError(RuntimeError):
    pass


class ProtocolError(RuntimeError):
    pass


class ManagementError(ProtocolError):
    def __init__(self, msg: str, script: str, out: str, err: str):
        super().__init__(self, msg)
        self.script = script
        self.out = out
        self.err = err
